#include "evalParamOfRaser.h"
#include "computePijComponent.h"
#include "likelihoodComputationSPvec.h"
#include "stochasticProcess.h"
#include "stochasticProcessSSRV.h"
#include "gammaDistribution.h"
#include "betaDistribution.h"
#include "bblEMSPvec.h"
#include "betaOmegaDistribution.h"


evalParamOfRaser::evalParamOfRaser(const tree& et,
				   const sequenceContainer& sc,
				   const vector<stochasticProcess*> &spVec,
				   paramName myparamName,
				   const distributionPlusInvariant * nuDistr,
				   const raserOptions::distrType nuDistrType,
				   bool isFixedNuCategories,
				   bool isEquallyProbableRateCategories,
				   suffStatGlobalGam *ssc)
    : _paramName(myparamName),_et(et),_sc(sc),_spVec(spVec),
				   _nuDistrType(nuDistrType), 
				   _isFixedNuCategories(isFixedNuCategories), 
				   _isEquallyProbableRateCategories(isEquallyProbableRateCategories),
				   _ssc(NULL) {
	if (!nuDistr)
		errorMsg::reportError("Error in evalParam::evalParam: input nu distribution is NULL");
	_nuDistr = nuDistr->clone();
	if (ssc) //pre-computed input up values
		_ssc = new suffStatGlobalGam(*ssc);
}

evalParamOfRaser::~evalParamOfRaser(){
  if (_nuDistr != NULL) delete _nuDistr;
  if (_ssc != NULL) delete _ssc;
}


evalParamOfRaser::evalParamOfRaser(const evalParamOfRaser &other): 
_et(other._et),_sc(other._sc),_spVec(other._spVec),_nuDistrType(other._nuDistrType),
_isFixedNuCategories(other._isFixedNuCategories),
_isEquallyProbableRateCategories(other._isEquallyProbableRateCategories),
_ssc(NULL)
{
	if (other._nuDistr == NULL)
		errorMsg::reportError("Error in evalParam copy constructor: other nu distribution is NULL");
	_nuDistr=other._nuDistr->clone();
	if (other._ssc)
		_ssc = new suffStatGlobalGam(*(other._ssc));
}


MDOUBLE evalParamOfRaser::operator()(MDOUBLE param){
	switch (_paramName){
		case alphaRateDistr:
			updateRateDistrAlpha(param);
			break;
		case alphaNuDistr:
			updateNuDistrAlpha(param);
			break;
		case betaNuDistr:
			updateNuDistrBeta(param);
			break;
		case probNu0:
			updateProbNu0(param);
			break;
		case probInfinity:
			updateProbInfinity(param);
			break;
		default: 
			errorMsg::reportError("evalParamOfRaser::operator, illegal parameter name");
			break;
	}
	
	MDOUBLE res(0.0);
	if (!_ssc)
		res = likelihoodComputationSPvec::getTreeLikelihoodAllPosAlphTheSame(_et,_sc,_spVec,_nuDistr);
	else
		res = likelihoodComputationSPvec::getTreeLikelihoodUpIsFilledSelectionGam(_et,_sc,_spVec,*_ssc,_nuDistr);
	
	return -res;	
}

void evalParamOfRaser::updateRateDistrAlpha(MDOUBLE param){
// update the alpha in each of the stochastic processes!
	for (int i = 0; i< _spVec.size(); ++i){
		replacementModelSSRV* pMulRM = static_cast<replacementModelSSRV*>(_spVec[i]->getPijAccelerator()->getReplacementModel());
		generalGammaDistribution* generalGammaDist = static_cast<generalGammaDistribution*>(pMulRM->getDistribution()); 
		generalGammaDist->setAlpha(param);
		if (!_isEquallyProbableRateCategories) {
			pMulRM->updateFreq();
		}
		pMulRM->updateQ();
	}
	if (_ssc) {
		computePijGam pi;
		pi._V.resize(_nuDistr->categories());
		for (int i=0; i < _spVec.size(); ++i) {
			pi._V[i].fillPij(_et,*_spVec[i]);
		}
		computeUpAlg cup;
		cup.fillComputeUp(_et,_sc,pi,*_ssc);
	}
}

void evalParamOfRaser::updateNuDistrAlpha(MDOUBLE param){
	switch (_nuDistrType){
		case raserOptions::generalGammaDist:	case raserOptions::gammaDist:
			static_cast<generalGammaDistribution*>(_nuDistr->getBaseDistribution())->setAlpha(param);
			break;
		case raserOptions::betaDist:
			static_cast<betaDistribution*>(_nuDistr->getBaseDistribution())->setAlpha(param);
			break;
		case raserOptions::betaInfDist:
			static_cast<betaOmegaDistribution*>(_nuDistr->getBaseDistribution())->setAlpha(param);
			break;
		case raserOptions::generalGammaInfDist: case raserOptions::gammaInfDist: {
			distributionPlusCategory* infDistCasted = static_cast<distributionPlusCategory*>(_nuDistr->getBaseDistribution());
			static_cast<generalGammaDistribution*>(infDistCasted->getBaseDistribution())->setAlpha(param);
			}
			break;
		default: errorMsg::reportError("evalParamOfRaser::updateNuDistrAlpha: nuDistrType is not supported"); 
			break;
	}
	if (!_isFixedNuCategories) {
		for (int i = 0; i < _nuDistr->categories(); ++i){
			MDOUBLE nu = _nuDistr->rates(i);
			static_cast<replacementModelSSRV*>(_spVec[i]->getPijAccelerator()->getReplacementModel())->setRateOfRate(nu);
		}
	}
}

void evalParamOfRaser::updateNuDistrBeta(MDOUBLE param){
	switch (_nuDistrType){
	case raserOptions::generalGammaDist:	case raserOptions::gammaDist:
		static_cast<generalGammaDistribution*>(_nuDistr->getBaseDistribution())->setBeta(param);
		break;
	case raserOptions::betaDist:
		static_cast<betaDistribution*>(_nuDistr->getBaseDistribution())->setBeta(param);
		break;
	case raserOptions::betaInfDist:
		static_cast<betaOmegaDistribution*>(_nuDistr->getBaseDistribution())->setBeta(param);
		break;
	case raserOptions::generalGammaInfDist: case raserOptions::gammaInfDist: {
		distributionPlusCategory* infDistCasted = static_cast<distributionPlusCategory*>(_nuDistr->getBaseDistribution());
		static_cast<generalGammaDistribution*>(infDistCasted->getBaseDistribution())->setBeta(param);
		}
		break;
	default: errorMsg::reportError("evalParamOfRaser::updateNuDistrBeta: nuDistrType is not supported"); 
		break;

	}
	if (!_isFixedNuCategories) {
		for (int i = 0; i < _nuDistr->categories(); ++i){
			MDOUBLE nu = _nuDistr->rates(i);
			static_cast<replacementModelSSRV*>(_spVec[i]->getPijAccelerator()->getReplacementModel())->setRateOfRate(nu);
		}
	}
}

void evalParamOfRaser::updateProbNu0(MDOUBLE param){
	_nuDistr->setInvProb(param);
}

void evalParamOfRaser::updateProbInfinity(MDOUBLE param){
	MDOUBLE probNu0 = _nuDistr->getInvProb();
	MDOUBLE relativeDistrPortion = 0.0;
	if (probNu0 != 1.0) {
		relativeDistrPortion = 1.0 - param / (1 - probNu0);
	}
	switch (_nuDistrType){
		case raserOptions::betaInfDist:
			static_cast<betaOmegaDistribution*>(_nuDistr->getBaseDistribution())->setBetaProb(relativeDistrPortion);
			break;
		case raserOptions::generalGammaInfDist: case raserOptions::gammaInfDist:
			static_cast<distributionPlusCategory*>(_nuDistr->getBaseDistribution())->setBaseDistProb(relativeDistrPortion);	
			break;
		default: errorMsg::reportError("evalParamOfRaser::updateProbInfinity: nuDistrType is not supported"); 
			break;
	}
}
