#include "raser.h"
#include "bblEM.h"
#include "likeDist.h"
#include "recognizeFormat.h"
#include "someUtil.h"
#include "alphabet.h"
#include "nj.h"
#include "likelihoodComputation.h"
#include "numRec.h"
#include "evaluateCharacterFreq.h"
#include "readDatMatrix.h"
#include "amino.h"
#include "nucleotide.h"
#include "likelihoodComputationSPvec.h"
#include "matrixUtils.h"
#include "evalParamOfRaser.h"
#include "treeUtil.h"
#include "bblEM.h"
#include "aaJC.h"
#include "nucJC.h"
#include "replacementModelSSRV.h"
#include "stochasticProcessSSRV.h"
#include "trivialAccelerator.h"
#include "uniDistribution.h"

#ifdef WIN32	
#include <process.h>
#else
#include <unistd.h>
#endif

using namespace std;

int main(int argc, char** argv)
{

	raser rs4s(argc,argv);
	return 0;
}


/***********constructor***************/
raser::raser(int argc, char** argv) :
_pOptions(NULL),
_baseSc(NULL),
_scSSRV(NULL),
_raserModel(NULL),
_alph(NULL),
_baseAlph(NULL),
_refSeq(NULL)
{
	time_t t1;
	time(&t1);
	time_t t2;

	printInfo(cout);
	fillOptionsParameters(argc,argv);
	
	myLog::setLog(_pOptions->_logFile, _pOptions->_verboseLevel);
	myLog::printArgv(5,argc,argv);
	int pp = getpid();
	LOG(5,<<"#Process_id= "<<pp<<endl);
	_pOptions->printOptionParameters();
	
	run(argc, argv);

	time(&t2);
	LOG(5,<<endl<<"TOTAL RUNNING TIME = "<<(t2-t1)/60.0<<" minutes"<<endl);
	myLog::endLog();
}

/***********RUNS THE PROGRAM***************
1. Reads the seqeunce data and input tree 
2. Creates a vector of stochastic processs based on the substitution model
3. Optimizes the parameters of the model 
4. Computes posterior probability per site
5. For sites with prob(shift)>0.95, find the nodes with the highest posterior probaility.
***f******************************************************/
void raser::run(int argc, char** argv){

	initScAndTree(); // initialize sequence container and tree
	if (_pOptions->_onlyPrintNodeIDs) {
		ofstream out2(_pOptions->_outTreeFileWithBranchesNames.c_str());
		printTreeWithNodeIdBPStyle(out2);
		out2.close();
		return;
	}
	_raserModel = new raserModel(_pOptions); 
	optimizeParams();
	if (!_pOptions->_useNullModel){
		initPositionsData();
		fillPij();
		computePosteriorProbs();
		if (_t.getLeavesNum()>5){
			getNodesPosteriorProbs4significantSites();		
		}
		else { // if 5 or less taxa, there is no point in relating to significant internal branches (only one such branch)
			LOGnOUT(5,<<endl<<"Less than 5 taxa, significant branches are not computed"<<endl);
			cout<<endl<<"Less than 5 taxa, significant branches are not computed"<<endl;
		}
		printResults();
	}
}


/***********destructor***************/
raser::~raser() {
	if (_pOptions) delete _pOptions;
	if (_baseSc) delete _baseSc;
	if (_scSSRV) delete _scSSRV;
	if (_alph) delete _alph;
	if (_baseAlph) delete _baseAlph;
	if (_refSeq) delete _refSeq;
	if (_raserModel) delete _raserModel;
}

void raser::printInfo(ostream& out) {
	out<<endl;
	out<<" ======================================================================="<<endl;
	out<<" The RAte Shift EstimatoR (RASER) program "<<endl;
	out<<" Version: 2.0 Last updated July 21th 2010 "<<endl;
	out<<" For program support, please contact Osnat Penn or Adi Stern"<<endl;
	out<<" Osnat Penn: penn@post.tau.ac.il, Adi Stern: sternadi@post.tau.ac.il "<<endl;
	out<<" ======================================================================="<<endl;
	out<<endl;
}

void raser::fillOptionsParameters(int argc, char* argv[]) {
	string paramStr = argv[1];
	_pOptions = new raserOptions();
	_pOptions->initOptions(paramStr);
}

/******INITIALIZE SEQ CONTAINER & TREE*****
This function reads the codon-aligned file according
to specified genetic code (default - nuc. standard)
and reads either a user-tree file or creates a NJ tree 
******************************************/
void raser::initScAndTree(){
	switch (_pOptions->_alphabetSize) {
		case 4 : _baseAlph = new nucleotide;  break;
		case 20 : _baseAlph = new amino; break; 
		default: errorMsg::reportError("Error in raser::initCodonScAndTree, raser implements only amino and nucleotide alphabets");
	}
	ifstream in(_pOptions->_inSeqFile.c_str());
	_baseSc = new sequenceContainer(recognizeFormat::read(in, _baseAlph));
	/***getting reference (query) sequence before gaps are changed to missing data*/
	int id;
	if (_pOptions->_inQuerySeq.size()==0) id = _baseSc->placeToId(0);
	else id  = _baseSc->getId(_pOptions->_inQuerySeq,false);

	if (id == -1) {
		id = _baseSc->placeToId(0);
		LOG(5,<<"Query sequence not found"<<endl);
		LOG(5,<<"USING 1st sequence as query: $$$"<< (*_baseSc)[id].name()<<"$$$"<<endl);
	}
	_refSeq = new sequence((*_baseSc)[id]);

	_baseSc->changeGaps2MissingData(); 
	_alph = new mulAlphabet(_baseAlph,_pOptions->_numOfCategoriesForRateDistr);
	_scSSRV = new sequenceContainer(*_baseSc,_alph);
	in.close();

	if (_pOptions->_inTreeFile.size()!=0){ //here checking if there is a user tree
		LOG(5,<<"-------------------------"<<endl);
		LOG(5,<<"Reading user tree "<<endl<<endl;); 
		tree et(_pOptions->_inTreeFile);
		_t = et;
	}
	else {
		createTree();
	}
	//if (_pOptions->_estimateInitialBLS)
	//	estimateInitialBLs();
}


/***********CREATE NJ TREE********************************
This function creates a NJ tree with a simple theoretical SSRV model
*********************************************************/
void raser::createTree(){
	LOG(5,<<"-------------------------"<<endl);
	LOGnOUT(5,<<endl<<"Reconstructing NJ tree "<<endl<<endl;); 
	VVdouble distM;
	vector<string> names;
	names.resize(_baseSc->numberOfSeqs());
	distM.resize(_baseSc->numberOfSeqs());
	int i;
	for (i=0;i<distM.size();i++){
		distM[i].resize(distM.size());
		names[i]=((*_baseSc)[i].name());		
	}	
	// creating a simplistic model for computing ML distances for a NJ tree
	uniDistribution distr;
	replacementModel *baseRM = NULL;
	if (_pOptions->_alphabetSize == 4)
		baseRM = new nucJC;
	else if (_pOptions->_alphabetSize == 20)
		baseRM = new aaJC;
	else 
		errorMsg::reportError("raser::createTree: Only amino and nucleotide alphabets implemented in raser");
	replacementModel *ssrvM = new replacementModelSSRV(&distr,baseRM,1.0);
	pijAccelerator* pij = new trivialAccelerator(ssrvM);
	stochasticProcessSSRV sp(pij);
	delete pij;
	delete baseRM;
	delete ssrvM;
	likeDist lkDist(sp);
	for ( i=0;i<_baseSc->numberOfSeqs();i++){
		for(int j=i+1;j<_baseSc->numberOfSeqs();j++){
			distM[i][j]= lkDist.giveDistance((*_baseSc)[i],(*_baseSc)[j],NULL);
		}
	}
	NJalg nj;	
	_t = nj.computeTree(distM,names);
	if (_pOptions->_specificNodeForStochasticMapping != -1) {
		tree::nodeP mynode = _t.findNodeById(_pOptions->_specificNodeForStochasticMapping);
		if (!mynode) {
			errorMsg::reportError("the node id inputted for stochastic mapping does not exist in the tree");
		}
	}

}

/***********ESTMIMATE INITIAL BRANCH LENGTHS***************
Initial one iteration estimation of branch lenghts.
This is used in cases where we want to use sum of
branch lengths for fixed categories, but the input tree des not have 
good estimates of BLs. A simple ssrv model based on JC is used
*********************************************************/

//void raser::estimateInitialBLs(){
//	uniDistribution distr;
//	replacementModel *baseRM = NULL;
//	if (_pOptions->_alphabetSize == 4)
//		baseRM = new nucJC;
//	else if (_pOptions->_alphabetSize == 20)
//		baseRM = new aaJC;
//	else 
//		errorMsg::reportError("raser::createTree: Only amino and nucleotide alphabets implemented in raser");
//	replacementModel *ssrvM = new replacementModelSSRV(&distr,baseRM,1.0);
//	pijAccelerator* pij = new trivialAccelerator(ssrvM);
//	stochasticProcessSSRV sp(pij);
//	delete pij;
//	delete baseRM;
//	delete ssrvM;
//	bblEM bbl(_t,*_baseSc,sp,NULL,1);
//	LOG(5,<<"Tree after initial estimation of branch lengths: "<<endl);
//	LOGDO(5,_t.output(myLog::LogFile()));
//
//}




/***********ML over the parameters***************
Finds best parameters
*********************************************************/

void raser::optimizeParams(){
	if (_pOptions->_manyStartPointsOpt) {
		Vint pointsNum(3,0);
		pointsNum[0] =5; pointsNum[1]=3; pointsNum[2]=1;
		Vint iterNum(3,0);
		iterNum[0] =0; iterNum[1]=3; iterNum[2]=7;
		Vdouble tols(3,_pOptions->_epsilonLikelihood);
		
		_raserModel->optimizeManyStart(_t,*_scSSRV,_pOptions,pointsNum,iterNum,tols);
	}
	else {
		_raserModel->optimizeParameters(_t,*_scSSRV,_pOptions,_pOptions->_numBBLiterations,_pOptions->_epsilonLikelihood);
	}
	_t.output(_pOptions->_outTreeFile);
	ofstream out(_pOptions->_outResFile.c_str());
	_raserModel->printModelParam(out,_pOptions);
	out.close();
}


/******************************************************
Print the results to the output file
*******************************************************/
void raser::printResults(){
	ofstream out(_pOptions->_outResFile.c_str(),ios::app);
	out.precision(2);
	out<<"Rate-Shifting Positions"<<endl;
	out<<"Displayed on sequence "<<_refSeq->name()<<endl;
	out<<"=========================================================================================================================================="<<endl;
	out<<"POS\t"<<"AMINO\t"<<"POSTERIOR PROBABILITY OF RATE-SHIFT (* IF > "<<_pOptions->_posteriorCutoffSites<<")"<<endl;
	distribution* nuDistr = _raserModel->getNuDistr();
	//print to log file actual values of each category
	LOG(5,<<endl<<"POS\t"<<"AMINO\t"<<"POSTERIOR PROBABILITIES OF CATEGORIES"<<endl<<"\t\t\t\t");
	for (int cat = 0; cat <nuDistr->categories(); ++cat){
		string cat2print = double2string(nuDistr->rates(cat));
		if (cat==nuDistr->categories()-1)//last category is 0
			cat2print = "0";
		if ((_pOptions->_nuDistrType == raserOptions::betaInfDist) || 
			(_pOptions->_nuDistrType == raserOptions::gammaInfDist) || 
			(_pOptions->_nuDistrType == raserOptions::generalGammaInfDist)){
				if (cat==nuDistr->categories()-2)//one before last category is 20, which stands for infinity
					cat2print = "Infinity";
			}
			LOG(5,<<cat2print<<"\t");
	}
	LOG(5,<<endl<<"=========================================================================================================================================="<<endl);

	int gapVal = _alph->gap();
	vector<posData>::iterator posDataItr;
	for (posDataItr=_positionsData.begin();posDataItr!=_positionsData.end();++posDataItr) {
		int posContent = posDataItr->getRefSeqContent();
		if (posContent==gapVal)
			continue;
		posDataItr->printPostVal(out,_baseAlph,_pOptions->_posteriorCutoffSites);
	}
	out.close();

	ofstream out2(_pOptions->_outTreeFileWithBranchesNames.c_str());
	printTreeWithNodeIdBPStyle(out2);
	out2.close();

	ofstream out3(_pOptions->_outNodesResFile.c_str());

	vector<posData>::iterator positionsDataItr;
	out3 << "pos" << '\t' << "node" << '\t' << "probAcc" << '\t' << "probDec" << '\t' << "probShift" << endl;
	for (positionsDataItr = _positionsData.begin(); positionsDataItr != _positionsData.end();++positionsDataItr) {
		positionsDataItr->printNodesData(out3,_pOptions->_numBranchesToPrint,_pOptions->_posteriorCutoffNodes);
	}
	out3.close();

}

void raser::fillPij() {
	distribution* nuDistr = _raserModel->getNuDistr();
	int numOfCategories = nuDistr->categories();
	_pij._V.resize(numOfCategories);
	vector<stochasticProcess*> spVec = _raserModel->getSpVec();
	LOG(5,<<endl<<"Prior nu distribution:"<<endl);
	for (int i=0; i < numOfCategories; ++i) {
		LOG(5,<<"Nu =  "<<nuDistr->rates(i)<<" Prob(nu) = "<<nuDistr->ratesProb(i)<<endl);
		_pij._V[i].fillPij(_t,*spVec[i]);
	}

}

/*********Compute raser (nu) values*************
Computes the Empirical Bayesian expectation of the posterior
estimators for nu values
fills the vector _positionsData
*******************************************************/
void raser::computePosteriorProbs(){
	distribution* nuDistr = _raserModel->getNuDistr();
	int numOfCategories = nuDistr->categories();
	
	LOGnOUT(5,<<endl<<"Calculating posterior and expectation of posterior values for all sites"<<endl);

	vector<posData>::iterator posDataItr;
	for (posDataItr=_positionsData.begin();posDataItr!=_positionsData.end();++posDataItr) {
		int pos_in_MSA = posDataItr->getPosInMSA();
		LOG(5,<<"Likelihood of position "<<pos_in_MSA<<endl);
		computeNu4site(pos_in_MSA, *posDataItr);
		posDataItr->calcPostValRateShift(numOfCategories-2);//assumes the last 2 categories are nu=0 & nu=inf
	}
}

void raser::initPositionsData() {
	int noOfgaps=0;
	int gapVal = _alph->gap();
	for (int pos_in_MSA=0; pos_in_MSA < _baseSc->seqLen(); ++pos_in_MSA) {
		int curPosContent = (*_refSeq)[pos_in_MSA];
		int pos_in_refSeq(-1);
		if (curPosContent==gapVal) {
			noOfgaps++;
		} else {
			pos_in_refSeq = pos_in_MSA-noOfgaps;
		}
		posData pos_data(pos_in_MSA,pos_in_refSeq,curPosContent);
		_positionsData.push_back(pos_data);
	}
}

void raser::computeNu4site(const int pos, posData& pos_data){
	distribution* nuDistr = _raserModel->getNuDistr();
	vector<stochasticProcess*> spVec = _raserModel->getSpVec();
	// here we compute the posterior P(nu|data) = P(nu)*P(data|nu) / P(data)
	VdoubleRep dblRepPosteriorV(nuDistr->categories(),0.0); // temporary dblRep container for posterior values
	doubleRep dblRepTotalLikelihood(0.0);// temporary dblRep container for total likelihood
	for (int i=0; i < nuDistr->categories(); ++i) {
	    dblRepPosteriorV[i] = likelihoodComputation::getLofPos(pos,_t,*_scSSRV,_pij[i],*spVec[i]);
	    LOG(5,<<"  Likelihood of category "<<i<<" (nu="<<nuDistr->rates(i)<<") = "<<convert(dblRepPosteriorV[i])<<endl);
		dblRepPosteriorV[i] *= nuDistr->ratesProb(i);
	    dblRepTotalLikelihood+=dblRepPosteriorV[i];
	}
	if (!(dblRepTotalLikelihood!=0))
		errorMsg::reportError("Error in raser::computeNu4site, likelihood = 0");

	// here we compute the expectation of the posterior: sigma nu * P(nu | data)
	for (int j=0; j < nuDistr->categories(); ++j) {
		dblRepPosteriorV[j]/=dblRepTotalLikelihood;
		pos_data.addPosteriorProb(convert(dblRepPosteriorV[j])); // revert back to DOUBLE
	}
}


void raser::getNodesPosteriorProbs4significantSites()
{
	distribution* nuDistr = _raserModel->getNuDistr();
	vector<stochasticProcess*> spVec = _raserModel->getSpVec();
	shiftNodes4site nodes4site(_t,*_scSSRV,*_baseSc,spVec,nuDistr,*_refSeq);
	if (_pOptions->_specificNodeForStochasticMapping == -1) {
		nodes4site.computeSignificantBranches(_positionsData,_pOptions->_posteriorCutoffSites,_pij);
	}
	else {
		tree::nodeP myNode = _t.findNodeById(_pOptions->_specificNodeForStochasticMapping);
		nodes4site.computeSignificantBranches(_positionsData,0.0,_pij,myNode);

	}
};


// copied from the original covarion code , from the file checkov.cpp
void raser::printTreeWithNodeIdBPStyle(ostream &out) const{
	recursivePrintTree(out,_t.getRoot());
	out<<";";
}

// similar to the file checkov.cpp from the original covarion code
// The bootstrap values is the nodes id.
void raser::recursivePrintTree(ostream &out,const tree::nodeP &myNode) const {
	if (myNode->isLeaf()) {
		out << myNode->name() << "_" << myNode->id();
		out << ":"<< myNode->dis2father();
		return;
	} else {
		out <<"(";
		for (int i=0;i<myNode->getNumberOfSons();++i) {
			if (i>0) out <<",";
			recursivePrintTree(out, myNode->getSon(i));
		}
		out <<")";
		if (myNode->isRoot()==false) {
			out<<":"<< myNode->dis2father();
			out << "["<<myNode->id()<<"]";
		}
	}
}
