#include "raserModel.h"
#include "betaDistribution.h"
#include "gammaDistributionFixedCategories.h"
#include "gammaDistributionLaguerre.h"
#include "betaDistribution.h"
#include "betaOmegaDistribution.h"
#include "distributionPlusInvariant.h"
#include "distributionPlusCategory.h"
#include "trivialAccelerator.h"
#include "chebyshevAccelerator.h"
#include "readDatMatrix.h"
#include "replacementModelSSRV.h"
#include "stochasticProcessSSRV.h"
#include "aaJC.h"
#include "nucJC.h"
#include "bblEMSPvec.h"
#include "likelihoodComputationSPvec.h"
#include "computePijComponent.h"
#include "evalParamOfRaser.h"
#include "talRandom.h"
#include <algorithm>

raserModel::raserModel(raserOptions* pOptions)
{
	createNuDistr(pOptions);
	createSpVec(pOptions);
	_bestRateDistrAlpha = static_cast<generalGammaDistribution*>
		(_spVec[0]->distr())->getAlpha();
	init(pOptions);
}


raserModel::raserModel(const raserModel &other) :
_bestRateDistrAlpha(other._bestRateDistrAlpha),
_bestNuDistrAlpha(other._bestNuDistrAlpha),
_bestNuDistrBeta(other._bestNuDistrBeta),
_bestProbNu0(other._bestProbNu0),
_bestProbInfinity(other._bestProbInfinity),
_bestL(other._bestL),
_lowerValueOfParamAlpha(other._lowerValueOfParamAlpha),
_upperValueOfParamAlpha(other._upperValueOfParamAlpha),
_lowerValueOfParamNu(other._lowerValueOfParamNu),
_upperValueOfParamNu(other._upperValueOfParamNu),
_lowerBoundProbNu0(other._lowerBoundProbNu0),
_upperBoundProbNu0(other._upperBoundProbNu0),
_epsilonForBrent(other._epsilonForBrent),
_maxBblIterations(other._maxBblIterations),
_maxTotalIterations(other._maxTotalIterations),
_isEquallyProbableRateCategories(other._isEquallyProbableRateCategories)
{
	_nuDistr = other._nuDistr->clone();
	_spVec.resize(other._spVec.size(),NULL);
	for (int i = 0; i< other._spVec.size(); ++i){
		if (other._spVec[i])
			_spVec[i] = other._spVec[i]->clone();
	}
}


raserModel::~raserModel()
{
	if (_nuDistr) delete _nuDistr;
	for (int i = 0; i< _spVec.size(); ++i){
		if (_spVec[i])
			delete _spVec[i];
	}
}


void raserModel::copy(const raserModel& other)
{
	_bestRateDistrAlpha=other._bestRateDistrAlpha;
	_bestNuDistrAlpha=other._bestNuDistrAlpha;
	_bestNuDistrBeta=other._bestNuDistrBeta;
	_bestProbNu0=other._bestProbNu0;
	_bestProbInfinity=other._bestProbInfinity;
	_bestL=other._bestL;
	_lowerValueOfParamAlpha=other._lowerValueOfParamAlpha;
	_upperValueOfParamAlpha=other._upperValueOfParamAlpha;
	_lowerValueOfParamNu=other._lowerValueOfParamNu;
	_upperValueOfParamNu=other._upperValueOfParamNu;
	_lowerBoundProbNu0=other._lowerBoundProbNu0;
	_upperBoundProbNu0=other._upperBoundProbNu0;
	_epsilonForBrent=other._epsilonForBrent;
	_maxBblIterations=other._maxBblIterations;
	_maxTotalIterations=other._maxTotalIterations;
	_isEquallyProbableRateCategories=other._isEquallyProbableRateCategories;

	if (_nuDistr) delete (_nuDistr);
	_nuDistr = other._nuDistr->clone();
	for (int i = 0; i< _spVec.size(); ++i){
		if (_spVec[i]) delete (_spVec[i]);
	}
	_spVec.clear();
	_spVec.resize(other._spVec.size(),NULL);
	for (int i = 0; i< other._spVec.size(); ++i){
		if (other._spVec[i])
			_spVec[i] = other._spVec[i]->clone();
	}
}


void raserModel::init(raserOptions* pOptions) {

	_upperValueOfParamAlpha = 2.5;
	_upperValueOfParamNu = 2.5;
	_epsilonForBrent = pOptions->_epsilonLikelihood;
	_maxBblIterations = pOptions->_numBBLiterations;
	_maxTotalIterations = 10;
	_isEquallyProbableRateCategories = true;
	if (pOptions->_useFixedCategoriesForRateDistribution ||
		pOptions->_useLaguerreRateDistribution)
		_isEquallyProbableRateCategories = false;

	_lowerValueOfParamAlpha = 0.1;
	_lowerValueOfParamNu = 0.1;
	_bestRateDistrAlpha = pOptions->_initAlphaRateDistr;
	_bestNuDistrAlpha = pOptions->_initAlphaNuDistr;
	_bestNuDistrBeta = pOptions->_initBetaNuDistr;
	_bestProbInfinity = pOptions->_initInfinityProb;
	_bestProbNu0 = pOptions->_initProbNu0;
	_lowerBoundProbNu0 = 0; 
	_upperBoundProbNu0 = 1.0-_lowerBoundProbNu0;

}


void raserModel::createNuDistr(raserOptions* pOptions) {
	
	distribution* baseNuDistr = createBaseNuDistr(pOptions);
	_nuDistr = new distributionPlusInvariant(baseNuDistr,pOptions->_initProbNu0);
	if (baseNuDistr) delete baseNuDistr;
}


distribution* raserModel::createBaseNuDistr(raserOptions* pOptions) {
	if (pOptions->_useNullModel)
		pOptions->_initInfinityProb = 1.0 - pOptions->_initProbNu0; // this means the probability of the gamma portion = 0
	//MDOUBLE relativeInitInfProb = pOptions->_initInfinityProb / (1-pOptions->_initProbNu0); //relative proportion within baseNuDistr
	// if _initProbNu0 = 1.0 there is no meaning for the rest of the probabilities.
	MDOUBLE relativeDistrPortion = 0.0;
	if (pOptions->_initProbNu0 != 1.0){
		relativeDistrPortion = 1.0-pOptions->_initInfinityProb / (1-pOptions->_initProbNu0); //relative proportion within baseNuDistr
	}
	MDOUBLE initAlphaForNuDistr = pOptions->_initAlphaNuDistr;
	MDOUBLE initBetaForNuDistr = pOptions->_initBetaNuDistr;
	distribution* baseNuDistr = NULL;
	if (!pOptions->_useFixedCategoriesForNuDistribution) // use equal-prob categories for the nu-distribution
	{
		switch (pOptions->_nuDistrType) {
			case (raserOptions::gammaDist) :
				baseNuDistr = new gammaDistribution(initAlphaForNuDistr,pOptions->_numOfCategoriesForNuDistr);
				break;
			case (raserOptions::generalGammaDist) :
				baseNuDistr = new generalGammaDistribution(initAlphaForNuDistr,initBetaForNuDistr,pOptions->_numOfCategoriesForNuDistr);
				break;
			case (raserOptions::betaDist):
				baseNuDistr = new betaDistribution(initAlphaForNuDistr,initBetaForNuDistr,pOptions->_numOfCategoriesForNuDistr);
				break;
			case (raserOptions::betaInfDist):
				if (pOptions->_infinityNu <= 1)
					errorMsg::reportError("The infinity category must be higher than 1");
				baseNuDistr = new betaOmegaDistribution(initAlphaForNuDistr,initBetaForNuDistr,pOptions->_numOfCategoriesForNuDistr,relativeDistrPortion,pOptions->_infinityNu);
				break;

			case (raserOptions::gammaInfDist): {
				distribution* withinInfDistr = new gammaDistribution(initAlphaForNuDistr,pOptions->_numOfCategoriesForNuDistr);
				baseNuDistr = new distributionPlusCategory(withinInfDistr,relativeDistrPortion,pOptions->_infinityNu);
				if (withinInfDistr) delete withinInfDistr;}
								 break;
			case (raserOptions::generalGammaInfDist):{
				distribution* withinInfDistr = new generalGammaDistribution(initAlphaForNuDistr,initBetaForNuDistr,pOptions->_numOfCategoriesForNuDistr);
				baseNuDistr = new distributionPlusCategory(withinInfDistr,relativeDistrPortion,pOptions->_infinityNu);
				if (withinInfDistr) delete withinInfDistr;}
									   break;
			default: errorMsg::reportError("raser::createStochasticProcessVec nuDistrType is not supported"); 
				break;
		}
	}
	else { //use fixed categories for the nu distribution
		    int numNuCategories = pOptions->_numOfCategoriesForNuDistr;
		    int numIntervals = numNuCategories+1;
		    
		    Vdouble fixedBoundaries(numIntervals,0.0);
		    fixedBoundaries[0]=0.0;
		    fixedBoundaries[numNuCategories] = VERYBIG/10000.0; //last boundary; to avoid overflow divided by 10000
		    MDOUBLE boundariesIntervalOfRateDist =  0.4;// done to preserve results in Penn et al. 2008 
		    for (int i = 1; i< numIntervals; ++i)
			fixedBoundaries[i] = i * boundariesIntervalOfRateDist;


		switch (pOptions->_nuDistrType) {
			case (raserOptions::gammaDist) :
//				baseNuDistr = new gammaDistributionFixedCategories(initAlphaForNuDistr,pOptions->_numOfCategoriesForNuDistr);
			    baseNuDistr = new gammaDistributionFixedCategories(fixedBoundaries,initAlphaForNuDistr);
				break;
			case (raserOptions::generalGammaDist) :
//				baseNuDistr = new generalGammaDistributionFixedCategories(initAlphaForNuDistr,initBetaForNuDistr,pOptions->_numOfCategoriesForNuDistr);
			    baseNuDistr = new generalGammaDistributionFixedCategories(fixedBoundaries,initAlphaForNuDistr,initBetaForNuDistr);
				break;
			case (raserOptions::gammaInfDist): {
//				distribution* withinInfDistr = new gammaDistributionFixedCategories(initAlphaForNuDistr,pOptions->_numOfCategoriesForNuDistr);
				distribution* withinInfDistr = new gammaDistributionFixedCategories(fixedBoundaries,initAlphaForNuDistr);
				baseNuDistr = new distributionPlusCategory(withinInfDistr,relativeDistrPortion,pOptions->_infinityNu);
				if (withinInfDistr) delete withinInfDistr; }
								 break;
			case (raserOptions::generalGammaInfDist): {
//				distribution* withinInfDistr = new generalGammaDistributionFixedCategories(initAlphaForNuDistr,initBetaForNuDistr,pOptions->_numOfCategoriesForNuDistr);
				distribution* withinInfDistr = new generalGammaDistributionFixedCategories(fixedBoundaries,initAlphaForNuDistr,initBetaForNuDistr);
				baseNuDistr = new distributionPlusCategory(withinInfDistr,relativeDistrPortion,pOptions->_infinityNu);
				if (withinInfDistr) delete withinInfDistr;}
										break;
			default: errorMsg::reportError("raser::createStochasticProcessVec, under fixed categories nuDistrType is not supported"); 
				break;
		}
	}
	return baseNuDistr;
}


void raserModel::createSpVec(raserOptions* pOptions) {
	_spVec.resize(_nuDistr->categories());
	for (int categor=0; categor<_nuDistr->categories();categor++){
		MDOUBLE nu = _nuDistr->rates(categor);
		replacementModel *baseRM = NULL;// base model, without SSRV.
		MDOUBLE initAlphaForRateDistr = pOptions->_initAlphaRateDistr;
		distribution *rateDistr = NULL;
		if (pOptions->_useFixedCategoriesForRateDistribution) {
		    //use fixed categories for the rate distribution
		    int numRateCategories = pOptions->_numOfCategoriesForRateDistr;
		    int numIntervals = numRateCategories+1;
		    
		    Vdouble fixedBoundaries(numIntervals,0.0);
		    fixedBoundaries[0]=0.0;
		    fixedBoundaries[numRateCategories] = VERYBIG/10000.0; //last boundary; to avoid overflow divided by 10000
		    MDOUBLE boundariesIntervalOfRateDist =  0.5;// done to preserve results in Penn et al. 2008 
		    for (int i = 1; i< numIntervals; ++i)
			fixedBoundaries[i] = i * boundariesIntervalOfRateDist;

//		    rateDistr = new gammaDistributionFixedCategories(initAlphaForRateDistr,pOptions->_numOfCategoriesForRateDistr); 
		    rateDistr = new gammaDistributionFixedCategories(fixedBoundaries,initAlphaForRateDistr); 
		} else if (pOptions->_useLaguerreRateDistribution) {
			// use Guass-Laguerre quadrature method for assessing rate distribution
			rateDistr = new gammaDistributionLaguerre(initAlphaForRateDistr,pOptions->_numOfCategoriesForRateDistr);
		} else { 
			// use equal-prob categories for the rate-distribution {
			rateDistr = new gammaDistribution(initAlphaForRateDistr,pOptions->_numOfCategoriesForRateDistr);			
		}

		pijAccelerator *basePijAcc=NULL; // Accelerator for the base model, without SSRV.
		switch (pOptions->_modelName) {
		case (raserOptions::day): 
			baseRM=new pupAll(datMatrixHolder::dayhoff);
			basePijAcc = new chebyshevAccelerator(baseRM);
			break;
		case (raserOptions::jtt):
			baseRM = new pupAll(datMatrixHolder::jones);
			basePijAcc = new chebyshevAccelerator(baseRM);
			break;
		case (raserOptions::rev):
			baseRM=new pupAll(datMatrixHolder::mtREV24);
			basePijAcc = new chebyshevAccelerator(baseRM);
			break;
		case (raserOptions::wag):
			baseRM=new pupAll(datMatrixHolder::wag);
			basePijAcc = new chebyshevAccelerator(baseRM);
			break;
		case (raserOptions::cprev):
			baseRM=new pupAll(datMatrixHolder::cpREV45);
			basePijAcc = new chebyshevAccelerator(baseRM);
			break;
		case (raserOptions::HIVb):
			baseRM = new pupAll(datMatrixHolder::HIVb); 
			basePijAcc = new chebyshevAccelerator(baseRM);
			break;
		case (raserOptions::HIVw):
			baseRM = new pupAll(datMatrixHolder::HIVw); 
			basePijAcc = new chebyshevAccelerator(baseRM);
			break;
		case (raserOptions::nucjc):
			baseRM = new nucJC;
			basePijAcc = new trivialAccelerator(baseRM);
			break;
		case (raserOptions::aajc):
			baseRM=new aaJC;
			basePijAcc = new trivialAccelerator(baseRM);
			break;
		default: errorMsg::reportError("raser::createStochasticProcessVec modelName is not supported");
		}
		replacementModel *repM = new replacementModelSSRV(rateDistr,baseRM,nu);
		pijAccelerator *pijAcc = new trivialAccelerator(repM);
		stochasticProcess *sp = new stochasticProcessSSRV(pijAcc);	
	
		_spVec[categor] = sp->clone();
		delete rateDistr;
		delete baseRM;
		delete basePijAcc;
		delete repM;
		delete pijAcc;
		delete sp;
	}
	

}


MDOUBLE raserModel::compLogLikelihood(const tree &et, const sequenceContainer &sc) {
	return likelihoodComputationSPvec::getTreeLikelihoodAllPosAlphTheSame
		(et,sc,_spVec,_nuDistr);
}


void raserModel::optimizeParameters(tree& et, const sequenceContainer& sc,
									  raserOptions* pOptions, int maxBblIterations /* = 3 */,
									  MDOUBLE epsilonForBrent /* = 0.1 */,
									  int maxTotalIterations /* = 10 */)
{

	_maxBblIterations = maxBblIterations;
	_epsilonForBrent = epsilonForBrent;
	_maxTotalIterations = maxTotalIterations;

	distributionPlusInvariant* nuDistrCasted = 
		static_cast<distributionPlusInvariant*>(_nuDistr);
	//initial likelihood value
	_bestL = compLogLikelihood(et,sc);
	LOG(5,<<"Likelihood initialized at = "<<_bestL<<endl);

	MDOUBLE newL = _bestL;
	MDOUBLE alphaRateDistrFound(0), alphaNuDistrFound(0), betaNuDistrFound(0), 
		curProbNu0(0), probInfinityFound(0);


	bool changed = false;
	bool wasChangedAlphaRateOrBBL = false; // if there is a change in alpha of rate or in BLs, up has to be re-computed for alpha of nu and prob nu=0

	LOGnOUT(5,<<endl<<"Beginning optimization of parameters"<<endl);
	LOG(5,<<"Parameters initialized to:"<<endl);
	LOG(5,<<"Alpha rate distribution = "<<_bestRateDistrAlpha<<endl);
	LOG(5,<<"Alpha nu distribution = "<<_bestNuDistrAlpha<<endl);
	if ((pOptions->_nuDistrType != raserOptions::gammaDist) && (pOptions->_nuDistrType != raserOptions::gammaInfDist))
		LOG(5,<<"Beta nu distribution = "<<_bestNuDistrBeta<<endl);
	if ((pOptions->_nuDistrType == raserOptions::betaInfDist) || (pOptions->_nuDistrType == raserOptions::gammaInfDist) 
		|| (pOptions->_nuDistrType == raserOptions::generalGammaInfDist))
		LOG(5,<<"Prob(nu=infinity) = "<<_bestProbInfinity<<endl);
	if (!pOptions->_fixedProbNu0)
		LOG(5,<<"Prob(nu=0)= "<<_bestProbNu0<<endl<<endl);

	suffStatGlobalGam *ssc = NULL; // this is used only if the nu categories are fixed. In this case up is pre-computed
	int iter = 0;
	for (iter=0; iter < _maxTotalIterations; ++iter) {
		LOG(5,<<endl<<"Iteration Number= " << iter <<endl);
		LOG(5,<<"---------------------"<<endl);		
		alphaNuDistrFound = alphaRateDistrFound = 0;
		changed = false;
		/****alpha of rate distribution****/
		if (!pOptions->_fixedAlphaRateDistr) {
			LOG(5,<<"Optimizing alpha of rate distribution" <<endl);
			newL = -brent(_lowerValueOfParamAlpha,
				_bestRateDistrAlpha,
				_upperValueOfParamAlpha,
				evalParamOfRaser(et,sc,_spVec,
				evalParamOfRaser::alphaRateDistr,nuDistrCasted,
				pOptions->_nuDistrType,pOptions->_useFixedCategoriesForNuDistribution,
				_isEquallyProbableRateCategories), //note there is no direct need for giving the nuDsitr here, but eh constructor requires it
				_epsilonForBrent,&alphaRateDistrFound); 
			LOG(5,<<"current best L = "<<_bestL<<endl<<endl);

			if (newL > _bestL+pOptions->_epsilonLikelihood) {
				_bestL = newL;
				_bestRateDistrAlpha = alphaRateDistrFound;
				for (int i = 0; i< _spVec.size(); ++i){
					replacementModelSSRV* pMulRM = 
						static_cast<replacementModelSSRV*>
						(_spVec[i]->getPijAccelerator()
							->getReplacementModel());
					generalGammaDistribution* generalGammaDist = static_cast<generalGammaDistribution*>(pMulRM->getDistribution()); 
					generalGammaDist->setAlpha(alphaRateDistrFound);
					if (!_isEquallyProbableRateCategories) {
						pMulRM->updateFreq();
					}
					pMulRM->updateQ();
					if (i==0) {
						LOG(6,<<endl<<"new rate distribution:"<<endl);
						for (int j = 0; j < generalGammaDist->categories(); ++j)
							LOG(6,<<"Rate = " << generalGammaDist->rates(j) << "\tprob(rate) = "
								<<generalGammaDist->ratesProb(j)<<endl);
					}
				}
				changed = true;
				wasChangedAlphaRateOrBBL = true;
				LOG(5,<<"new L After alpha of rate Distr = " <<newL<<endl);
				LOG(5,<<"new alpha accepted" <<endl);
				LOG(5,<<"alpha of rate Distr is now = " <<_bestRateDistrAlpha<<endl<<endl);
			} 
			else {// reverting alpha back to old value, since spVec alpha value was changed by evalParamOfRaser
				for (int i = 0; i< _spVec.size(); ++i){
					replacementModelSSRV* pMulRM = static_cast<replacementModelSSRV*>(_spVec[i]->getPijAccelerator()->getReplacementModel());
					generalGammaDistribution* generalGammaDist = static_cast<generalGammaDistribution*>(pMulRM->getDistribution()); 
					generalGammaDist->setAlpha(_bestRateDistrAlpha);
					if (pOptions->_useFixedCategoriesForRateDistribution) {
						pMulRM->updateFreq();
					}
					pMulRM->updateQ();
				}
				LOG(5,<<"new alpha rejected" <<endl);
				LOG(5,<<"Proposed likelihood was " <<newL<<" for alpha = "<<alphaRateDistrFound<<endl<<endl);
			}
		}

		/****alpha of nu distribution****/
		bool doOptimizeAlphaNu = true;
		if (pOptions->_fixedProbNu0 && (nuDistrCasted->getInvProb()==1)) // prob (nu=0) = 1
			doOptimizeAlphaNu = false;
		if (pOptions->_useNullModel)
			doOptimizeAlphaNu = false;
		if (pOptions->_fixedAlphaNuDistr)
			doOptimizeAlphaNu = false;
		if (doOptimizeAlphaNu){
			LOG(5,<<"Optimizing alpha of nu distribution" <<endl);
			if (pOptions->_useFixedCategoriesForNuDistribution) { // pre-compute up values, since these will not change when categories are fixed
				if (ssc == NULL) {// if not filled yet (NULL), which means we are in the first iteration 
					ssc = new suffStatGlobalGam;
					wasChangedAlphaRateOrBBL = true; // this is in order to enter the next if
				}
				if (wasChangedAlphaRateOrBBL) {
					fillSuffStatGlobalGam(*ssc,et,sc);
					wasChangedAlphaRateOrBBL = false; // indicates no need to re-fill until next change
				}
			}
			newL = -brent(_lowerValueOfParamNu,
				_bestNuDistrAlpha,
				_upperValueOfParamNu,
				evalParamOfRaser(et,sc,_spVec,evalParamOfRaser::alphaNuDistr,
				nuDistrCasted,pOptions->_nuDistrType,pOptions->_useFixedCategoriesForNuDistribution,_isEquallyProbableRateCategories,ssc),
				_epsilonForBrent,&alphaNuDistrFound); 

			LOG(5,<<"current best L = "<<_bestL<<endl<<endl);
			if (newL > _bestL+pOptions->_epsilonLikelihood ) {
				_bestL = newL;
				_bestNuDistrAlpha = alphaNuDistrFound;
				switch (pOptions->_nuDistrType){
					case raserOptions::gammaDist: case raserOptions::generalGammaDist:
						static_cast<generalGammaDistribution*>(nuDistrCasted->getBaseDistribution())->setAlpha(alphaNuDistrFound);
						break;
					case raserOptions::betaDist:
						static_cast<betaDistribution*>(nuDistrCasted->getBaseDistribution())->setAlpha(alphaNuDistrFound);
						break;
					case raserOptions::betaInfDist:
						static_cast<betaOmegaDistribution*>(nuDistrCasted->getBaseDistribution())->setAlpha(alphaNuDistrFound);
						break;
					case raserOptions::gammaInfDist: case raserOptions::generalGammaInfDist: {
						distributionPlusCategory* infDistCasted = static_cast<distributionPlusCategory*>(nuDistrCasted->getBaseDistribution());
						static_cast<generalGammaDistribution*>(infDistCasted->getBaseDistribution())->setAlpha(alphaNuDistrFound);
									   }
									   break;
					default: errorMsg::reportError("raserModel::optimizeParameters: nuDistrType is not supported"); 
						break;
				}
				if (!pOptions->_useFixedCategoriesForNuDistribution) {
					for (int i = 0; i < nuDistrCasted->categories(); ++i){
						MDOUBLE nu = nuDistrCasted->rates(i);
						static_cast<replacementModelSSRV*>
							(_spVec[i]->getPijAccelerator()->getReplacementModel())->setRateOfRate(nu);
					}
				}
				changed = true;
				LOG(5,<<"new L After alpha of nu Distr = " <<newL<<endl);
				LOG(5,<<"new alpha accepted" <<endl);
				LOG(5,<<"alpha of nu Distr is now = " <<_bestNuDistrAlpha<<endl<<endl);
				LOG(6,<<endl<<"new nu distribution:"<<endl);
				for (int j = 0; j < nuDistrCasted->categories(); ++j){
					LOG(6,<<"nu = " << nuDistrCasted->rates(j) << "\tprob(nu) = "
						<<nuDistrCasted->ratesProb(j)<<endl);
				}
			} else {
				LOG(5,<<"new alpha rejected" <<endl);
				LOG(5,<<"Proposed likelihood was " <<newL<<" for alpha = "<<alphaNuDistrFound<<endl<<endl);
			}
		}

		/****beta of nu distribution****/
		bool doOptimizeBetaNu = true;
		if (pOptions->_nuDistrType == raserOptions::gammaDist)
			doOptimizeBetaNu = false;
		if (pOptions->_fixedProbNu0 && (nuDistrCasted->getInvProb()==1)) // prob (nu=0) = 1
			doOptimizeBetaNu = false;
		if (pOptions->_fixedBetaNuDistr)
			doOptimizeBetaNu = false;
		if (pOptions->_useNullModel)
			doOptimizeBetaNu = false;
		if (doOptimizeBetaNu){
			LOG(5,<<"Optimizing beta of nu distribution" <<endl);
			if (pOptions->_useFixedCategoriesForNuDistribution) { // pre-compute up values, since these will not change when categories are fixed
				if (ssc == NULL) {// if not filled yet (NULL), which means we are in the first iteration 
					ssc = new suffStatGlobalGam;
					wasChangedAlphaRateOrBBL = true; // this is in order to enter the next if
				}
				if (wasChangedAlphaRateOrBBL) {
					fillSuffStatGlobalGam(*ssc,et,sc);
					wasChangedAlphaRateOrBBL = false; // indicates no need to re-fill until next change
				}
			}
			newL = -brent(_lowerValueOfParamNu,
				_bestNuDistrBeta,
				_upperValueOfParamNu,
				evalParamOfRaser(et,sc,_spVec,evalParamOfRaser::betaNuDistr,
				nuDistrCasted,pOptions->_nuDistrType,pOptions->_useFixedCategoriesForNuDistribution,_isEquallyProbableRateCategories,ssc),
				_epsilonForBrent,&betaNuDistrFound); 

			LOG(5,<<"current best L = "<<_bestL<<endl<<endl);
			if (newL > _bestL+ pOptions->_epsilonLikelihood) {
				_bestL = newL;
				_bestNuDistrBeta = betaNuDistrFound;
				switch (pOptions->_nuDistrType){
					case raserOptions::gammaDist: case raserOptions::generalGammaDist:
						static_cast<generalGammaDistribution*>(nuDistrCasted->getBaseDistribution())->setBeta(betaNuDistrFound);
						break;
					case raserOptions::betaDist:
						static_cast<betaDistribution*>(nuDistrCasted->getBaseDistribution())->setBeta(betaNuDistrFound);
						break;
					case raserOptions::betaInfDist:
						static_cast<betaOmegaDistribution*>(nuDistrCasted->getBaseDistribution())->setBeta(betaNuDistrFound);
						break;
					case raserOptions::gammaInfDist: case raserOptions::generalGammaInfDist: {
						distributionPlusCategory* infDistCasted = static_cast<distributionPlusCategory*>(nuDistrCasted->getBaseDistribution());
						static_cast<generalGammaDistribution*>(infDistCasted->getBaseDistribution())->setBeta(betaNuDistrFound);
									   }
									   break;
					default: errorMsg::reportError("raserModel::optimizeParameters: nuDistrType is not supported"); 
						break;
				}
				if (!pOptions->_useFixedCategoriesForNuDistribution) {
					for (int i = 0; i < nuDistrCasted->categories(); ++i){
						MDOUBLE nu = nuDistrCasted->rates(i);
						static_cast<replacementModelSSRV*>(_spVec[i]->getPijAccelerator()->getReplacementModel())->setRateOfRate(nu);
					}
				}
				changed = true;
				LOG(5,<<"new L After beta of nu Distr = " <<newL<<endl);
				LOG(5,<<"new beta accepted" <<endl);
				LOG(5,<<"beta of nu Distr is now = " <<_bestNuDistrBeta<<endl<<endl);
				for (int j = 0; j < nuDistrCasted->categories(); ++j){
					LOG(6,<<"nu = " << nuDistrCasted->rates(j) << "\tprob(nu) = "
						<<nuDistrCasted->ratesProb(j)<<endl);
				}
			} 
			else {
				LOG(5,<<"new beta of nu Distr rejected" <<endl);
				LOG(5,<<"Proposed likelihood was " <<newL<<" for beta = "<<betaNuDistrFound<<endl<<endl);
			}
		}

		/****probability of nu=0 ****/
		if (!pOptions->_fixedProbNu0){
			LOG(5,<<"Optimizing prob(nu=0)" <<endl);
			if (pOptions->_useFixedCategoriesForNuDistribution) { // pre-compute up values, since these will not change when categories are fixed
				if (ssc == NULL) {// if not filled yet (NULL), which means we are in the first iteration and alpha of nu was fixed
					ssc = new suffStatGlobalGam;
					wasChangedAlphaRateOrBBL = true; // this is only in order to enter the next if
				}
				if (wasChangedAlphaRateOrBBL) 
					fillSuffStatGlobalGam(*ssc,et,sc);
				wasChangedAlphaRateOrBBL = false; // indicates no need to re-fill until next change

			}
			newL = -brent(_lowerBoundProbNu0,
				_bestProbNu0,
				_upperBoundProbNu0,
				evalParamOfRaser(et,sc,_spVec,evalParamOfRaser::probNu0,nuDistrCasted,
				pOptions->_nuDistrType,pOptions->_useFixedCategoriesForNuDistribution,_isEquallyProbableRateCategories,ssc),
				_epsilonForBrent,&curProbNu0); 

			LOG(5,<<"current best L = "<<_bestL<<endl<<endl);
			if (newL > _bestL+pOptions->_epsilonLikelihood ) {
				_bestL = newL;
				if (_bestProbNu0 != 1.0) { // previous prob(nu=0) value
					// set prob infinity according to its new relative proportion
					_bestProbInfinity = (1.0-curProbNu0)*_bestProbInfinity/(1.0-_bestProbNu0);
					setProbNu0AndInfinity(curProbNu0, _bestProbInfinity,pOptions->_nuDistrType);
				}
				else { // ** if prob nu 0 in the previous iteration=1.0, no need to re-compute relative values
					_bestProbInfinity = getRelativeProbNuInf(pOptions->_nuDistrType)*(1-curProbNu0);
					nuDistrCasted->setInvProb(curProbNu0);
				}
				_bestProbNu0 = curProbNu0;

				changed = true;
				LOG(5,<<"new L After prob(nu=0) = " <<newL<<endl);
				LOG(5,<<"new prob(nu=0) accepted" <<endl);
				LOG(5,<<"prob(nu=0) is now = " <<_bestProbNu0<<endl<<endl);
				LOG(5,<<"following this change prob(nu=inf) is now = " <<_bestProbInfinity<<endl<<endl);
				for (int j = 0; j < nuDistrCasted->categories(); ++j){
					LOG(6,<<"nu = " << nuDistrCasted->rates(j) << "\tprob(nu) = "
						<<nuDistrCasted->ratesProb(j)<<endl);
				}
			} 
			else {
				LOG(5,<<"new prob(nu=0) rejected" <<endl);
				LOG(5,<<"Proposed likelihood was " <<newL<<" for prob(nu=0) = "<<curProbNu0<<endl<<endl);
			}
			
			// check the upper boundary likelihood ;
			LOG(5,<<endl<<"checking the upper boundary of prob(nu=0)"<<endl);
			evalParamOfRaser eval_bound(et,sc,_spVec,evalParamOfRaser::probNu0,nuDistrCasted,
				pOptions->_nuDistrType,pOptions->_useFixedCategoriesForNuDistribution,_isEquallyProbableRateCategories,ssc);
			MDOUBLE bound_1_L = -eval_bound(_upperBoundProbNu0);
			if (bound_1_L > _bestL+ pOptions->_epsilonLikelihood ) {
				_bestL = bound_1_L;
				// set prob infinity 
				_bestProbInfinity = 1-_upperBoundProbNu0;
				_bestProbNu0 = _upperBoundProbNu0;

				nuDistrCasted->setInvProb(_bestProbNu0);
				changed = true;
				LOG(5,<<"new L After prob(nu=0) = " <<_bestL<<endl);
				LOG(5,<<"new prob(nu=0) accepted" <<endl);
				LOG(5,<<"prob(nu=0) is now = " <<_bestProbNu0<<endl<<endl);

			}
			else {
			    LOG(5,<<"upper bound ("<<_upperBoundProbNu0<<") of prob(nu=0) rejected" <<endl);
			}
			

		}
		/****probability of nu=infinity ****/
		bool doOptimizeInfProb = true;
		if (pOptions->_useNullModel)
			doOptimizeInfProb = false;
		if (pOptions->_fixedInfinityProb)
			doOptimizeInfProb = false;
		if (((pOptions->_nuDistrType == raserOptions::betaDist)||(pOptions->_nuDistrType == raserOptions::gammaDist)|| (pOptions->_nuDistrType == raserOptions::generalGammaDist))){
			doOptimizeInfProb = false;
		}
		if (_bestProbNu0 == 1.0){
			doOptimizeInfProb = false;
		}
		if (doOptimizeInfProb){
			LOG(5,<<"Optimizing prob(nu=infinity)" <<endl);
			MDOUBLE initGuess = _bestProbInfinity;
			if (initGuess>(1.0-_bestProbNu0)){ // if the initial guess is larger than the upper bound, which is defined by prob(nu=0)
			    initGuess = ((1.0-_bestProbNu0)-_lowerBoundProbNu0)/2.0;
			}
			newL = -brent(_lowerBoundProbNu0,
				initGuess,
				(1.0-_bestProbNu0), // the upper bound of probInf is 1-probNu0
				evalParamOfRaser(et,sc,_spVec,evalParamOfRaser::probInfinity,nuDistrCasted,
				pOptions->_nuDistrType,pOptions->_useFixedCategoriesForNuDistribution,_isEquallyProbableRateCategories,ssc),
				_epsilonForBrent,&probInfinityFound); 

			LOG(5,<<"current best L = "<<_bestL<<endl<<endl);
			if (newL > _bestL+pOptions->_epsilonLikelihood ) {
				_bestL = newL;
				_bestProbInfinity = probInfinityFound;
				MDOUBLE relativeProbInf = _bestProbInfinity / (1 - _bestProbNu0);
				setRelativeProbNuInf(relativeProbInf,pOptions->_nuDistrType);
				changed = true;
				LOG(5,<<"new L After prob(infinity) = " <<newL<<endl);
				LOG(5,<<"new prob(infinity) accepted" <<endl);
				LOG(5,<<"prob(infinity) is now = " <<probInfinityFound<<endl<<endl);
				for (int j = 0; j < nuDistrCasted->categories(); ++j){
					LOG(6,<<"nu = " << nuDistrCasted->rates(j) << "\tprob(nu) = "
						<<nuDistrCasted->ratesProb(j)<<endl);
				}

			} 
			else {
				LOG(5,<<"new prob(infinity) rejected" <<endl);
				LOG(5,<<"Proposed likelihood was " <<newL<<" for prob(infinity) = "<<probInfinityFound<<endl<<endl);
			}

			// check the upper boundary likelihood (prob(inifinity)=1-probNu0
			LOG(5,<<endl<<"checking upper boundary of prob(infinity)"<<endl);
			evalParamOfRaser eval_bound(et,sc,_spVec,evalParamOfRaser::probInfinity,nuDistrCasted,
				pOptions->_nuDistrType,pOptions->_useFixedCategoriesForNuDistribution,_isEquallyProbableRateCategories,ssc);
			MDOUBLE upperBoundNuInf = 1.0-_bestProbNu0;
			MDOUBLE up_bound_L = -eval_bound(upperBoundNuInf);
			if (up_bound_L > _bestL+ pOptions->_epsilonLikelihood ) {
					_bestL = up_bound_L;
					_bestProbInfinity = 1.0-_bestProbNu0;
					MDOUBLE relativeProbInf = 1.0; // if the upper bound is reached, this means that the beta portion iz zero and prob nu=inf is 1 relatively
				setRelativeProbNuInf(relativeProbInf,pOptions->_nuDistrType);
				
				changed = true;
				LOG(5,<<"new L After prob(infinity) = " <<_bestL<<endl);
				LOG(5,<<"new prob(infinity) accepted" <<endl);
				LOG(5,<<"prob(infinity) is now = " <<_bestProbInfinity<<endl<<endl);

			}
			else {
			    LOG(5,<<"upper bound ("<<1.0-_bestProbNu0<<") of prob(nu=0) rejected" <<endl);
			}
		}

		/****BBL***/
		if (pOptions->_bblOpt==true) {
			//using epsilonBrent as the epsilon for pairwise distances here		
			bblEMspVec bbl(et,sc,_spVec,_nuDistr,NULL,_maxBblIterations, pOptions->_epsilonLikelihood,
				_epsilonForBrent);
			newL = bbl.getTreeLikelihood();
			LOG(5,<<"current best L= "<<_bestL<<endl<<endl);
			LOG(5,<<"new L After BBL = " << newL<<endl);
			LOG(5,<<"Tree after this BBL iteration: "<<endl);
			LOGDO(5,et.output(myLog::LogFile()));
			if (newL > _bestL+pOptions->_epsilonLikelihood) {
				_bestL = newL;
				changed = true;
				wasChangedAlphaRateOrBBL = true;
			} else {
				if (newL < _bestL) {
					errorMsg::reportError("Error in raserModel::optimizeParameters: likelihood went down in bbl");
				} else {
					_bestL = newL;
				}
			}	
		}
		if (changed==false) {
			//perform another round of bbl em to make sure converged
			if (pOptions->_bblOpt==true) {
				LOG(5,<<"Testing if bbl converged"<<endl);
				bblEMspVec bbl(et,sc,_spVec,_nuDistr,NULL,_maxBblIterations,pOptions->_epsilonLikelihood,
					_epsilonForBrent);
				newL = bbl.getTreeLikelihood();
				LOG(5,<<"current best L= "<<_bestL<<endl<<endl);
				LOG(5,<<"new L After BBL = " << newL<<endl);
				LOG(5,<<"Tree after this BBL iteration: "<<endl);
				LOGDO(5,et.output(myLog::LogFile()));
				if (newL > _bestL+pOptions->_epsilonLikelihood) {
					_bestL = newL;
					changed = true;
					wasChangedAlphaRateOrBBL = true;
				}
				else {
					if (newL < _bestL) {
						errorMsg::reportError("Error in raserModel::optimizeParameters: likelihood went down in bbl");
					} else {
						_bestL = newL;
					}
					break;
				}
			} 
			else
				break;
		}
	}

	LOGnOUT(5,<<endl<<"Finished optimization of parameters for this round"<<endl<<endl);
	if (ssc)
		delete ssc;

	if (iter==_maxTotalIterations) {
		LOG(5,<<"Reached maximal number of iterations ("<<_maxTotalIterations<<") in raserModel::optimizeParameters. The last optimized parameters are used for the calculations."<<endl<<endl);
	}

}


void raserModel::fillSuffStatGlobalGam(suffStatGlobalGam &ssc, const tree &et,
													  const sequenceContainer &sc)
{
	computePijGam pi;
	pi._V.resize(_nuDistr->categories());
	for (int i=0; i < _spVec.size(); ++i) {
		pi._V[i].fillPij(et,*_spVec[i]);
	}
	computeUpAlg cup;
	cup.fillComputeUp(et,sc,pi,ssc);
}


void raserModel::printModelParam(ofstream& outFile,raserOptions* pOptions) const
{
	outFile<<"RASER Results File";
	if (pOptions->_useNullModel)
		outFile<<" for Null Model Run (ASRV+homogenous)";
	outFile<<endl;
	outFile<<"============================================================"<<endl;
	outFile<<"Parameters are:"<<endl;
	outFile<<"Log-likelihood: "<<getBestL()<<endl<<
		"Alpha for rate distribution: "<<getBestRateDistrAlpha()<<endl<<
		"Prob (nu=0): "<<getBestProbNu0()<<endl;
	if (!pOptions->_useNullModel){
		outFile<<"Alpha for nu distribution: "<<getBestNuDistrAlpha()<<endl;
		if (pOptions->_nuDistrType != raserOptions::gammaDist)
			outFile<<"Beta for nu distribution: "<<getBestNuDistrBeta()<<endl;
	}
	if ((pOptions->_nuDistrType == raserOptions::betaInfDist) || (pOptions->_nuDistrType == raserOptions::gammaInfDist) || (pOptions->_nuDistrType == raserOptions::generalGammaInfDist))
		outFile<<"Prob (nu=infinity): "<<_bestProbInfinity<<endl;
	LOG(5,<<endl<<"Prior rate distribution:"<<endl);
	//outFile<<endl<<"Prior rate distribution:"<<endl;
	replacementModelSSRV* pMulRM = static_cast<replacementModelSSRV*>(_spVec[0]->getPijAccelerator()->getReplacementModel());
	generalGammaDistribution* generalGammaDist = static_cast<generalGammaDistribution*>(pMulRM->getDistribution()); 
	for (int i = 0; i < generalGammaDist->categories(); ++i) {
		LOG(5,<<"Rate = " << generalGammaDist->rates(i) << "\tprob(rate) = "<<generalGammaDist->ratesProb(i)<<endl);
		//outFile <<"Rate = " << generalGammaDist->rates(i) << "\tprob(rate) = "<<generalGammaDist->ratesProb(i)<<endl;
	}
}


//// pointsNum, iterNum and tols are vectors of the same size in which each entry represent one cycle of optimization.
//// First, pointsNum[0] random starting points are created. 
//// Then, in each cycle i, the likelihood is calculated for pointsNum[i] points and iterNum[i] 
//// optimization iterations are done for each point with epsilons taken from tols[i]
//// We continue to the next cycle with the best pointsNum[i+1] points only.
void raserModel::optimizeManyStart(tree &et, const sequenceContainer &sc, raserOptions* pOptions, const Vint pointsNum, const Vint iterNum, const Vdouble tols) {

	int numOfOptCycles = pointsNum.size();
	LOG(3, <<"RUNNING MULTIPLE STARTING POINT OPTIMIZATION ("<<numOfOptCycles<<"CYCLES)"<<endl;);
	LOG(3, <<"First, "<<numOfOptCycles<<" of random points are created and LL is calculated for each"<<endl;);
	LOG(3, <<"In each cycle, a subset of the best points is chosen, till the last cycle where the best point is used and fully optimized"<<endl;);
	checkStartPointsInput(pointsNum,iterNum,tols);
	vector<raserModel*> modelVec;
	vector<tree*> treeVec;
	fillStartingModels(et,sc,pointsNum[0],pOptions,modelVec,treeVec);
	
	//make a small number of iterations for all random starts 
	Vdouble likelihoodVec;
	for (int i = 0; i < numOfOptCycles; ++i)
	{
	    LOG(3, <<"       ++++++Cycle " <<i<<"++++++"<<endl<<endl;);
		if (i != 0)
		{
		    vector<raserModel*> tmpModelVec(0);
		    vector<tree*> tmpTreeVec(0);
		    //sort results and continue optimization only with the best (pointsNum[i]) points
		    Vdouble sortedL = likelihoodVec;
		    sort(sortedL.begin(),sortedL.end());
		    MDOUBLE threshold = sortedL[sortedL.size()- pointsNum[i]];
		    for (int j = 0; j < likelihoodVec.size(); ++j)
		    {
			if (likelihoodVec[j] >= threshold) {
			    tmpModelVec.push_back(modelVec[j]);
			    tmpTreeVec.push_back(treeVec[j]);
			}
			else {
			    delete modelVec[j];
			    delete treeVec[j];
			}
		    }
		    modelVec.clear();
		    modelVec = tmpModelVec;
		    treeVec.clear();
		    treeVec = tmpTreeVec;
		} 

		likelihoodVec.clear();
		likelihoodVec.resize(pointsNum[i]); 
		int c; 		
		for (c = 0; c < pointsNum[i]; ++c)
		{
		    LOG(3, <<"  =====optimizing point " <<c<<" ======"<<endl;);
		    modelVec[c]->optimizeParameters(*(treeVec[c]), sc,pOptions,_maxBblIterations,tols[i],iterNum[i]);
		    MDOUBLE ll = modelVec[c]->getBestL();
		    LOG(3, <<"point: "<<c<<"  likelihood = "<<ll<<endl<<endl;);
		    likelihoodVec[c] = ll;
		}
	}
	Vdouble sortedL = likelihoodVec;
	sort(sortedL.begin(),sortedL.end());
	MDOUBLE bestL = sortedL[likelihoodVec.size() - 1];
	LOG(3, <<endl<<"FINAL LIKELIHOODS++++++++++++++"<<endl;);
	for (int i = 0; i < likelihoodVec.size(); ++i)
	{
	    if (bestL == likelihoodVec[i]) 
	    {
		copy(*(modelVec[i]));
		et = *(treeVec[i]);
	    }

	    LOG(3, <<"point "<<i<<" likelihood = "<<likelihoodVec[i]<<endl;);
	}	
	
	clearModelVec(modelVec,treeVec);
}


void raserModel::checkStartPointsInput(const Vint& pointsNum, const Vint& iterNum, 
													 const Vdouble& tols)
{

	//make sure that the number of points in each cycle is not bigger than the previous cycle.
	for (int i = 0; i < pointsNum.size()-1; ++i) {
		if (pointsNum[i] < pointsNum[i+1])
			errorMsg::reportError("input error in raserModel::checkStartPointsInput");
		}

	//make sure that pointNum, iterNum and tols are of the same size
	int pointNumSize = pointsNum.size();
	int iterNumSize = iterNum.size();
	int tolsSize = tols.size();
	if ((pointNumSize!=iterNumSize) || (iterNumSize!=tolsSize))
		errorMsg::reportError("input error in raserModel::checkStartPointsInput");
}

void raserModel::fillStartingModels(tree& et, const sequenceContainer& sc, int numStartingPoints
									  ,raserOptions* pOptions, vector<raserModel*>& modelVec,
									  vector<tree*>& treeVec)
{
	clearModelVec(modelVec,treeVec);
	modelVec.resize(numStartingPoints);
	treeVec.resize(numStartingPoints);

	for (int i = 0; i < numStartingPoints; ++i)
	{
		treeVec[i] = new tree(et);
		//the first model is identical to the current one
		if (i == 0) {
			modelVec[i] = new raserModel(*this);
		} 
		else {
			// create random model
			raserOptions randomOptions(*pOptions);
			randomOptions._initAlphaRateDistr = 
				talRandom::giveRandomNumberBetweenTwoPoints(_lowerValueOfParamAlpha, 
					_upperValueOfParamAlpha/2.0);
			randomOptions._initAlphaNuDistr = 
				talRandom::giveRandomNumberBetweenTwoPoints(_lowerValueOfParamNu, 
					_upperValueOfParamNu/2.0);
			randomOptions._initBetaNuDistr = 
				talRandom::giveRandomNumberBetweenTwoPoints(_lowerValueOfParamNu, 
					_upperValueOfParamNu/2.0);
			randomOptions._initProbNu0 = 
				talRandom::giveRandomNumberBetweenTwoPoints(_lowerBoundProbNu0, 
				_upperBoundProbNu0);
			if (i==1) {// create point which is the null model
				randomOptions._initInfinityProb = 1.0-randomOptions._initProbNu0;
				randomOptions._useNullModel = true;
				LOG(6,<< "++The following random point is the null model:"<<endl);
			}
			else { // not the null model
				randomOptions._initInfinityProb = 
					talRandom::giveRandomNumberBetweenTwoPoints(_lowerBoundProbNu0, 
						1.0-randomOptions._initProbNu0); // probInf is between 0.05 and probNu0
			}
			LOG(6,<< "create random point number " << i << endl << "Parameters are:" << 
				" -a " << randomOptions._initAlphaNuDistr << 
				" -c " << randomOptions._initBetaNuDistr <<
				" -d " << randomOptions._initAlphaRateDistr <<
				" -p " << randomOptions._initProbNu0 <<
				" -w " << randomOptions._initInfinityProb << endl;)
			modelVec[i] = new raserModel(&randomOptions);
		}
	}
}


void raserModel::clearModelVec(vector<raserModel*>& modelVec,vector<tree*>& treeVec)
{
	for(int i=0; i < modelVec.size(); ++i)
		if (modelVec[i]) delete (modelVec[i]);
	modelVec.clear();
	for(int i=0; i < treeVec.size(); ++i)
		if (treeVec[i]) delete (treeVec[i]);
	treeVec.clear();
}



void raserModel::setProbNu0AndInfinity(MDOUBLE newValProbNu0, MDOUBLE newValProbInfinity,raserOptions::distrType distType){
	MDOUBLE oldValProbInfinity(-1.0), oldValProbNu0(-1.0);
	distributionPlusInvariant* nuDistrCasted = 
		static_cast<distributionPlusInvariant*>(_nuDistr);
	oldValProbNu0 = nuDistrCasted->getInvProb();
	oldValProbInfinity=getRelativeProbNuInf(distType);
	
	nuDistrCasted->setInvProb(newValProbNu0);
// compute prob infinity according to its new relative proportion
	MDOUBLE relativeProbInf;
	if (newValProbNu0 != 1.0) { // if = 1.0 no meaning to change probNuInf
		relativeProbInf = newValProbInfinity / (1.0 - newValProbNu0);
		setRelativeProbNuInf(relativeProbInf,distType);
	}
}

MDOUBLE raserModel::getRelativeProbNuInf(raserOptions::distrType distType){
	MDOUBLE val = -1.0;
	distributionPlusInvariant* nuDistrCasted = 
		static_cast<distributionPlusInvariant*>(_nuDistr);
	switch (distType){
	case raserOptions::betaInfDist:
		val=static_cast<betaOmegaDistribution*>(nuDistrCasted->getBaseDistribution())->getBetaProb();
		break;
	case raserOptions::gammaInfDist: case raserOptions::generalGammaInfDist:
		val=static_cast<distributionPlusCategory*>(nuDistrCasted->getBaseDistribution())->getBaseDistProb();
		break;
	default: errorMsg::reportError("raserModel::getRelativeProbNuInf: nuDistrType is not supported "); 
		break;
	}
	return val;

}

void raserModel::setRelativeProbNuInf(MDOUBLE relativeProbNuInf,raserOptions::distrType distType){
	distributionPlusInvariant* nuDistrCasted = 
		static_cast<distributionPlusInvariant*>(_nuDistr);
	switch (distType){
	case raserOptions::betaInfDist:
		static_cast<betaOmegaDistribution*>(nuDistrCasted->getBaseDistribution())->setBetaProb(1.0-relativeProbNuInf);
		break;
	case raserOptions::gammaInfDist: case raserOptions::generalGammaInfDist:
		static_cast<distributionPlusCategory*>(nuDistrCasted->getBaseDistribution())->setBaseDistProb(1.0-relativeProbNuInf);
		break;
	default: errorMsg::reportError("raserModel::setRelativeProbNuInf: nuDistrType is not supported "); 
		break;
	}

}
