#ifndef __RASER_MODEL
#define __RASER_MODEL

#include "definitions.h"
#include "stochasticProcess.h"
#include "tree.h"
#include "sequenceContainer.h"
#include "gammaDistribution.h"
#include "talRandom.h"
#include "raserOptions.h"
#include "computeUpAlg.h"

/******************************************************
nu (the rate of rate param) is distributed under a  
discrete distribution. Thus, a vector of stochastic
processes is created for each mean value of the distribution
*********************************************************/
class raserModel{
public:
	raserModel(raserOptions* pOptions);   
	raserModel(const raserModel& other);
	void copy(const raserModel& other);
	virtual ~raserModel();
	
	//optimize the model param and return best log-likelihood
	void optimizeParameters(tree& et, 
		const sequenceContainer& sc,
		raserOptions* pOptions,
		int maxBblIterations = 3,MDOUBLE epsilonForBrent = 0.1, int maxTotalIterations = 10);
	

	void optimizeManyStart(tree &et, 
		const sequenceContainer &sc, 
		raserOptions* pOptions,
		const Vint pointsNum, 
		const Vint iterNum, 
		const Vdouble tols);


	MDOUBLE compLogLikelihood(const tree &et, const sequenceContainer &sc);
	
	void fillSuffStatGlobalGam(suffStatGlobalGam &ssc, const tree &et,  
		const sequenceContainer &sc);
	
	const MDOUBLE getBestRateDistrAlpha() const {return _bestRateDistrAlpha;}
	const MDOUBLE getBestNuDistrAlpha() const {return _bestNuDistrAlpha;}
	const MDOUBLE getBestNuDistrBeta() const {return _bestNuDistrBeta;}
	const MDOUBLE getBestProbNu0() const {return _bestProbNu0;}
	const MDOUBLE getBestProbInfinity() const {return _bestProbInfinity;}
	const MDOUBLE getBestL() const {return _bestL;}

	distribution* getNuDistr() {return _nuDistr;}
	vector<stochasticProcess*> getSpVec() {return _spVec;}
	void printModelParam(ofstream& outFile,raserOptions* pOptions) const;


private:
	void createNuDistr(raserOptions* pOptions);
	distribution* createBaseNuDistr(raserOptions* pOptions);
	void createSpVec(raserOptions* pOptions);
	void init(raserOptions* pOptions);
	

	void checkStartPointsInput(const Vint& pointsNum, const Vint& iterNum, const Vdouble& tols);
	void fillStartingModels(tree& et, const sequenceContainer& sc, int numStartingPoints
		,raserOptions* pOptions, vector<raserModel*>& modelVec,vector<tree*>& treeVec);
	void clearModelVec(vector<raserModel*>& modelVec,vector<tree*>& treeVec);

	void setProbNu0AndInfinity(MDOUBLE newValProbNu0, MDOUBLE newValProbInfinity,raserOptions::distrType distType); // compute the relative change in prob(infinity) following a chnage in prob nu0
	MDOUBLE getRelativeProbNuInf(raserOptions::distrType distType);
	void setRelativeProbNuInf(MDOUBLE relativeProbNuInf,raserOptions::distrType distType);

private:
	vector <stochasticProcess*> _spVec; //save stochasticProcess for each category 
	distribution * _nuDistr; // distribution over rate of rates

private:
	MDOUBLE _bestRateDistrAlpha;
	MDOUBLE _bestNuDistrAlpha;
	MDOUBLE _bestNuDistrBeta;
	MDOUBLE _bestProbNu0;
	MDOUBLE _bestProbInfinity;
	MDOUBLE _bestL;
	MDOUBLE _lowerValueOfParamAlpha;
	MDOUBLE _upperValueOfParamAlpha;
	MDOUBLE _lowerValueOfParamNu;
	MDOUBLE _upperValueOfParamNu;
	MDOUBLE _lowerBoundProbNu0;
	MDOUBLE _upperBoundProbNu0;
	MDOUBLE _epsilonForBrent;
	int _maxBblIterations;
	int _maxTotalIterations;
	bool _isEquallyProbableRateCategories;
};
#endif // __RASER_MODEL
