#ifndef ___RATESHIFT4SITE__OPTIONS_H
#define ___RATESHIFT4SITE__OPTIONS_H

#include "definitions.h"
#include <string>
#include <fstream>
using namespace std;

/*
--- utilize CLASS:Parameters ---
USAGE: SETTING DEFAULT PARAMETERS
Note that the type of the parameter is set according to the addParameter arguments. 
e.g., If a parameter is set using addParameter with an integer argument then subsequent updates (using updateParameter) 
to the same parameter will all be stored as integers. 
Therefore the following code should output a 0:
EXAMPLE
Parameters::addParameter("Dummy", 3);
Parameters::updateParameter("Dummy", "This should set it to zero");
cout << Parameters::getString("Dummy");
END
Note also that when setting default values of float parameters always use
a decimal point or else these parameters will be added as integers. 

USAGE: READING PARAMETERS FROM FILE
The readParameters method receives an input stream from which parameters are to be read. 
Files are structured so that each line specifies the value of a parameter. 
Each line gives the parameter name, a white space and then the parameter value. 
Lines whose first non white-space charachter is # are ignored. 
A basic schema for using the Parameters class is to set the default
values using addParameter calls and then calling readParameters to read in
parameters with other values or new parameters. 
EXAMPLE
Parameters::addParameter("CubeSize", 1.0);
Parameters::addParameter("MinVote", 8);
ifstream params("params");
Parameters::readParameters(params);
params.close();
Parameters::dump(cout);
END
With the following parameters file:
EXAMPLE
CubeSize 0.5
File  pdb4hhb.ent
END
The following output should result:
EXAMPLE
CubeSize (Float) 0.5
File     (Str)   pdb4hhb.ent
MinVote  (Int)   8
END

USAGE: SUBCLASSING AND PERFORMANCE
The Parameters engine keeps the parameters in a sorted list.
The correct usage would have been to inherit: e.g., class ProgParams : protected Parameters

*/

class raserOptions{
public:
	enum distrType{gammaDist,generalGammaDist,betaDist,betaInfDist,gammaInfDist,generalGammaInfDist};
	enum modelNameType {rev,jtt,day,aajc,nucjc,wag,cprev,HIVb,HIVw};

public:
	virtual ~raserOptions();

	void initOptions(const string& paramFileName);
	void initDefault();
	void readParameters(const string& paramFileName);
	void getParamsFromFile(const string& paramFileName);
	void getOutDirFromFile(const string& paramFileName);
	void verifyConsistParams();
	void printOptionParameters();

	// conversions from enum to (from) string 
	string getDistributionType(distrType type);
	distrType getDistributionType(const string& str);
	string getModelNameType(modelNameType type);
	modelNameType getModelNameType(const string& str);

public:
//################### Basic parameters:
// input (general)

    string _inSeqFile;
    string _inTreeFile;
	string _inQuerySeq;
    int _numBranchesToPrint;
	int _specificNodeForStochasticMapping;    

// output
	string _logFile;   
	int _verboseLevel; // verbose log level
    string _outTreeFile;
	string _outTreeFileWithBranchesNames;
    string _outResFile; //file with nu and posterior values for site
	string _outNodesResFile;
  
//################################################## Model params    
    modelNameType _modelName;
    distrType _nuDistrType;

    int _numOfCategoriesForRateDistr; // number of categories for rates distribution
    int _numOfCategoriesForNuDistr; // number of categories for nu distribution

	bool _onlyPrintNodeIDs; // if true, don't run raser at all.

	bool _fixedProbNu0;
    MDOUBLE _initProbNu0;// probability of nu=0, i.e prob. of no rate-shift
    bool _fixedAlphaRateDistr;
    MDOUBLE _initAlphaRateDistr;
    bool _fixedAlphaNuDistr;
    MDOUBLE _initAlphaNuDistr;
    bool _fixedBetaNuDistr;
    MDOUBLE _initBetaNuDistr;
    bool _fixedInfinityProb;
    MDOUBLE _initInfinityProb; // this is the relative probability within distribution+category (the distributions are nested)
    MDOUBLE _posteriorCutoffSites;
    MDOUBLE _posteriorCutoffNodes;
    

    bool _useFixedCategoriesForNuDistribution;
    bool _useFixedCategoriesForRateDistribution;
    bool _useLaguerreRateDistribution;
    bool _useNullModel;
//    MDOUBLE _upperBoundNuForFixedcategories;
//    MDOUBLE _upperBoundRateForFixedcategories;
    MDOUBLE _infinityNu;

    int _alphabetSize;
    bool _bblOpt; //true if branch lengths are to be optimized
    bool _manyStartPointsOpt; // true for many start points.
    MDOUBLE _epsilonLikelihood;
    int _numBBLiterations; // number of optimization iterations

};
#endif
